use crate::laser_diode::current_sources::{self, CurrentSourcePhy};
use crate::laser_diode::max5719;
use crate::thermostat::ad5680;
use crate::thermostat::max1968::{self, MAX1968PinSet, MAX1968Phy, PWM_FREQ_KHZ};
use stm32_eth::EthPins;
use stm32f4xx_hal::{
    gpio::{gpioa::*, gpiob::*, gpioc::*, gpiog::*, GpioExt, Input},
    otg_fs::USB,
    pac::{
        GPIOA, GPIOB, GPIOC, GPIOD, GPIOG, OTG_FS_DEVICE, OTG_FS_GLOBAL, OTG_FS_PWRCLK, SPI1, SPI2,
        TIM4,
    },
    rcc::Clocks,
    spi::{NoMiso, Spi},
    timer::pwm::PwmExt,
};

pub type EthernetPins =
    EthPins<PA1<Input>, PA7<Input>, PB11<Input>, PG13<Input>, PB13<Input>, PC4<Input>, PC5<Input>>;

pub fn setup(
    clocks: Clocks,
    tim4: TIM4,
    gpioa: GPIOA,
    gpiob: GPIOB,
    gpioc: GPIOC,
    gpiod: GPIOD,
    gpiog: GPIOG,
    spi1: SPI1,
    spi2: SPI2,
    otg_fs_global: OTG_FS_GLOBAL,
    otg_fs_device: OTG_FS_DEVICE,
    otg_fs_pwrclk: OTG_FS_PWRCLK,
) -> (
    EthernetPins,
    USB,
    CurrentSourcePhy<current_sources::Channel0>,
    MAX1968Phy<max1968::Channel0>,
    // thermostat_phy
) {
    let gpioa = gpioa.split();
    let gpiob = gpiob.split();
    let gpioc = gpioc.split();
    let gpiod = gpiod.split();
    let gpiog = gpiog.split();

    let usb = USB {
        usb_global: otg_fs_global,
        usb_device: otg_fs_device,
        usb_pwrclk: otg_fs_pwrclk,
        pin_dm: gpioa.pa11.into_alternate(),
        pin_dp: gpioa.pa12.into_alternate(),
        hclk: clocks.hclk(),
    };

    let eth_pins = EthPins {
        ref_clk: gpioa.pa1,
        crs: gpioa.pa7,
        tx_en: gpiob.pb11,
        tx_d0: gpiog.pg13,
        tx_d1: gpiob.pb13,
        rx_d0: gpioc.pc4,
        rx_d1: gpioc.pc5,
    };

    let current_source_phy = CurrentSourcePhy {
        dac: max5719::Dac::new(Spi::new(
                    spi2,
                    (
                        gpiob.pb10.into_alternate(),
                        NoMiso {},
                        gpiob.pb15.into_alternate(),
                    ),
                    max5719::SPI_MODE,
                    max5719::SPI_CLOCK_MHZ.convert(),
                    &clocks,
                ), gpiod.pd8.into_push_pull_output(),
                gpiob.pb14.into_push_pull_output(),
            ),
        current_source_ldo_en_pin: gpiod.pd9.into_push_pull_output(),
        current_source_short_pin: gpioa.pa4.into_push_pull_output(),
        pd_mon_pin: gpioa.pa3.into_analog()
    };


    let pwm_chs = (
        gpiob.pb6.into_alternate(),
        gpiob.pb7.into_alternate(),
        gpiob.pb8.into_alternate(),
    );
    let (max_i_neg0, max_v0, max_i_pos0) =
        tim4.pwm_hz(pwm_chs, PWM_FREQ_KHZ.convert(), &clocks).split();

    let max1968_phy = MAX1968Phy::new(MAX1968PinSet {
        dac: ad5680::Dac::new(
            Spi::new(
                spi1,
                (
                    gpiob.pb3.into_alternate(),
                    NoMiso {},
                    gpiob.pb5.into_alternate(),
                ),
                ad5680::SPI_MODE,
                ad5680::SPI_CLOCK_MHZ.convert(),
                &clocks,
            ),
            gpiob.pb4.into_push_pull_output(),
        ),
        shdn: gpioa.pa5.into_push_pull_output(),
        vref_pin: gpioa.pa6.into_analog(),
        itec_pin: gpiob.pb1.into_analog(),
        // Fixme: Flywire is added to Rev0_2 prototype. In Rev0_3, it is connected to PC0
        //dac_feedback_pin: gpioc.pc0.into_analog(),
        dac_feedback_pin: gpioc.pc3.into_analog(),
        vtec_pin: gpiob.pb0.into_analog(),
        max_v: max_v0,
        max_i_pos: max_i_pos0,
        max_i_neg: max_i_neg0,
    });

    (eth_pins, usb, current_source_phy, max1968_phy)
}
